@@ -34,6 +34,7 @@ import subprocess
 
 from datetime import datetime, timedelta
 from contextlib import contextmanager, closing
+import socket   # replace with ipaddress when this is py3-only
 
 from dsl_parser.constants import PLUGIN_INSTALL_KEY, PLUGIN_NAME_KEY
 
@@ -202,7 +203,7 @@ def get_manager_file_server_url():
     port = get_manager_rest_service_port()
     scheme = get_manager_file_server_scheme()
     return [
-        '{0}://{1}:{2}/resources'.format(scheme, host, port)
+        '{0}://{1}:{2}/resources'.format(scheme, ipv6_url_compat(host), port)
         for host in get_manager_rest_service_host()
     ]
 
@@ -1032,3 +1033,20 @@ def parse_schedule_datetime_string(date_str):
         raise NonRecoverableError(
             "{} is not a legal time format. accepted formats are "
             "YYYY-MM-DD HH:MM | HH:MM".format(date_str))
+
+
+def _is_ipv6(addr):
+    """Verifies if `addr` is a valid IPv6 address."""
+    # TODO replace socket with ipaddress once we're py3-only
+    try:
+        socket.inet_pton(socket.AF_INET6, addr)
+    except socket.error:
+        return False
+    return True
+
+
+def ipv6_url_compat(addr):
+    """Return URL-compatible version of IPv6 address (or just an address)."""
+    if _is_ipv6(addr):
+        return '[{0}]'.format(addr)
+    return addr
