@@ -25,7 +25,6 @@ from cloudify.manager import (
     get_rest_client,
     get_node_instance,
     update_execution_status,
-    update_node_instance
 )
 from cloudify.constants import (
     MGMTWORKER_QUEUE,
@@ -341,7 +340,6 @@ class WorkflowTask(object):
             # operation explicitly requested a retry, so we ignore
             # the handler set on the task.
             handler_result = HandlerResult.retry()
-            handler_result.retry_after = result.retry_after
         elif self.on_failure:
             handler_result = self.on_failure(self)
         else:
@@ -470,16 +468,14 @@ class RemoteWorkflowTask(WorkflowTask):
             self.set_state(TASK_SENDING)
         try:
             self._set_queue_kwargs()
-            task = self.workflow_context.internal.handler.get_task(
-                self, queue=self._task_queue, target=self._task_target,
-                tenant=self._task_tenant)
             self.workflow_context.internal.handler.wait_for_result(
-                self.async_result, self, task)
+                self, self._task_target)
             if should_send:
                 self.workflow_context.internal.send_task_event(
                     TASK_SENDING, self)
                 self.set_state(TASK_SENT)
-                self.workflow_context.internal.handler.send_task(self, task)
+                self.workflow_context.internal.handler.send_task(
+                    self, self._task_target, self._task_queue)
         except (exceptions.NonRecoverableError,
                 exceptions.RecoverableError) as e:
             self.set_state(TASK_FAILED)
@@ -977,10 +973,11 @@ class _SetNodeInstanceStateTask(_LocalTask):
         }
 
     def remote(self):
-        node_instance = get_node_instance(self._node_instance_id)
-        node_instance.state = self._state
-        update_node_instance(node_instance)
-        return node_instance
+        get_rest_client().node_instances.update(
+            self._node_instance_id,
+            force=True,
+            state=self._state
+        )
 
     def local(self):
         self.storage.update_node_instance(
